/**************************************************************************

  EXAMPLE2.C

  Written by Eric Jorgensen (1995)

  The is an example program to illustrate how to use Varmint's Audio
  Tools to play .mid files.


**************************************************************************/

#include "sound.h"
#include <conio.h>
#include <stdarg.h>


void vcprintf(char *string,...);
void mididemo(void);

extern WORD         dma_bufferlen;
extern int          sample_rate;
extern WORD         io_addr;
extern WORD         intnr;
extern WORD         dma_ch;
extern WORD         card_id;
extern WORD         fm_addr;
extern WORD         dsp_vers;
extern WORD         vsync_toolong;
extern CHANNEL      chan[4];
extern int          sounds_in_queue;
extern BYTE         defaultpatchmap[16];
extern int          mpu_timeout;
extern int          mpu_available;
extern char         vatdebugstring[];
extern VOICE        midi_voice[];
extern long int     debugnum;
extern int 					midi_callfreq;

BYTE mypatchmap[16] = {0,0,0,0,0,1,1,1,1,1,2,2,2,2,2,2};
																					// Initial FM patchmap to set
																					// tracks 1 and 2 to harpsichord
BYTE myfmpatchmap[32] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
													 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
char *instname[16] = {"ACGPIANO","ACPiano","ELGPIANO","HONKTONK","ELPIANO1",
				"ElPiano2","HARPSCHD","CLAVICHD","CELESTA","GLOCK","MUSICBOX",
				"VIBES","MARIMBA","XYLOPHON","TUBEBELL","Dulcimer"};
char errstring[256];
int mpu_checked = FALSE;


/**************************************************************************
	void main(int argc, char *argv[])

  DESCRIPTION:

**************************************************************************/
void main(int argc, char *argv[])
{
  int i;


                                           // You need a MIDI structure to
                                          // hold the data from the .mid file.
  MIDI far *song;
                                          // You will also need a character
                                          // string to catch any error
                                          // messages from LoadMidi().
  char errstring[100];
                                          // Lastly, you will need an
                                           // In general, it is good to perform
                                           // all file access before settup
                                           // up the sound blaster.
  if(argc < 2) exit(1);
  song = LoadMidi(argv[1],errstring);
  if(!song) {
    printf("ERROR loading midi:  %s\n",errstring);
    exit(1);
  }

  dma_bufferlen = 4;
	                                         // Set up the sound blaster.
  if(!SBSetUp()) {
    printf("SB_Setup returned this error: %s \n",errname[sberr]);
    FreeMidi(song);                       // Free the Midi you just loaded.
    exit(1);
  }

  GoVarmint();                            // Start up the sound Kernel.

  FMReset();                              // Get the FM chip ready.


  midi_data = song;                       // Tell VAT which song to play.
	mididemo();
  DropDeadVarmint();                      // Stop the SB interrupt.
  SBCleanUp();                            // Clean up sound stuff.
  FreeMidi(song);                         // Free our sond effect.

}

/**************************************************************************
	void mididemo(void)

	DESCRIPTION: Example function that loads and "plays" a  midi file

**************************************************************************/
void mididemo(void)
{
	int i;
	char r=0;

	clrscr();                                   // clear screen

	FMReset();

	midi_fmpatchmap = myfmpatchmap;

/*  for(i = 0 ; i < 9; i++) {                   // initialize voices
		FMSetVolume(i,0);
		FMSetVoice ( i,inst[0]) ;
	}  */



	textcolor(MAGENTA);
	gotoxy(30,2);
	vcprintf("MIDI demo menu");
	textcolor(CYAN);
	gotoxy(30,3);
	vcprintf("1   - Start music");
	gotoxy(30,4);
	vcprintf("2   - Stop music");
	gotoxy(30,5);
	vcprintf("3   - Pause music");
	gotoxy(30,6);
	vcprintf("4   - Resume music");
	gotoxy(30,7);
	vcprintf("5   - Rewind music");
	gotoxy(30,8);
	vcprintf("V,v - Change music volume");
	gotoxy(30,9);
	vcprintf("T,t - Change music tempo");
	gotoxy(30,10);
	vcprintf("F   - Toggle FM output: ");
	if(midi_fmout) vcprintf("ON  ");
	else vcprintf("OFF ");
	gotoxy(30,11);
	vcprintf("U   - Toggle MPU output: ");
	if(midi_mpuout) vcprintf("ON  ");
	else vcprintf("OFF ");
	gotoxy(30,12);
	vcprintf("P   - Set MPU-401 MIDI port");
	gotoxy(30,13);
	vcprintf("A   - Toggle MIDI patchmap ");
	if(midi_patchmap == defaultpatchmap) vcprintf("(16 channel)");
	else vcprintf("( 3 channel)");
	gotoxy(30,14);
	vcprintf("R   - Rotate FM patchmap (%s)",instname[myfmpatchmap[1]]);
	gotoxy(30,16);
	vcprintf("Q   - Quit");

	if(!mpu_available && mpu_checked) {
		gotoxy(1,20);
		vcprintf("Current MPU-401 midi port is invalid.");
	}


	while(toupper(r) != 'Q') {                  // MAIN LOOP
		gotoxy(30,18);                            // print temppo and volume stats
		textcolor(WHITE);
		vcprintf("Music Volume: %d  ",midi_volume);
		gotoxy(30,19);
		vcprintf("Music Tempo:  %d ",midi_usertempo);

		while(!kbhit()) {                         // handle the keyboard
			if(mpu_checked) {
				gotoxy(1,2);                          // check midi status while we wait
				if(mpu_timeout < 1000) vcprintf("MPU status: OK!        ");
				else vcprintf("MPU status: Timing out   ");
				gotoxy(1,3);
				vcprintf("Current port: %X  ",midi_mpuport);
			}
			for(i = 0; i < 9; i++) {
				gotoxy(1,5+i);
				if(midi_voice[i].active) textcolor(YELLOW);
				else textcolor(BROWN);
				vcprintf("Voice %d:  %d %d %d",i,midi_voice[i].note,
										midi_voice[i].owner_track,midi_voice[i].owner_channel);
			}
      gotoxy(40,1);
      vcprintf("(%d,%ld)",midi_callfreq,debugnum);
		}
		r = getch();                              // get keystroke

		if(r == 'V') {                            // Handle keystrokes
			midi_volume++;
			if(midi_volume > 0x3f) midi_volume = 0x3f;
		}
		else if(r == 'v') {
			if(midi_volume > 0) midi_volume--;
		}
		else if(r == '1') MidiCommand(v_play);
		else if(r == '2') MidiCommand(v_stop);
		else if(r == '3') MidiCommand(v_pause);
		else if(r == '4') MidiCommand(v_resume);
		else if(r == '5') MidiCommand(v_rewind);
		else if(toupper(r) == 'R') {              // Rorate FM patchmap
			myfmpatchmap[1]++;
			if(myfmpatchmap[1] > 15) myfmpatchmap[1] = 0;
			myfmpatchmap[2] = myfmpatchmap[1];
			gotoxy(55,14);
			vcprintf("(%s)     ",instname[myfmpatchmap[1]]);
		}
		else if(toupper(r) == 'A') {              // Toggle midi patchmap
			if(midi_patchmap == mypatchmap) midi_patchmap = defaultpatchmap;
			else midi_patchmap = mypatchmap;
			gotoxy(57,13);
			if(midi_patchmap == defaultpatchmap) vcprintf("(16 channel)");
			else vcprintf("( 3 channel)");
		}
		else if(r == 'T') {
			midi_usertempo *= 1.1;
			if(midi_usertempo > 1000) midi_usertempo = 1000;
		}
		else if(r == 't') {
			midi_usertempo *= 0.9;
			if(midi_usertempo < 10) midi_usertempo = 10;
		}
		else if(toupper(r) == 'P') {               // Get a new MPU port address
			gotoxy(1,22);
			vcprintf("Enter new port address in hex values [%x]: ",midi_mpuport);
			gets(errstring);
			if(strlen(errstring)) sscanf(errstring,"%x",&midi_mpuport);

																							 // check for accidents
			if(midi_mpuport < 0x200 || midi_mpuport > 0x360) midi_mpuport = 0x330;
			MPUEnter();
			gotoxy(1,23);
			if(!mpu_available) vcprintf("Current MPU-401 midi port is invalid.");
			else vcprintf("                                               ");
			gotoxy(1,22);
			vcprintf("                                                    ");
			mpu_checked = TRUE;
		}
		else if(toupper(r) == 'F') {               // Toggle output
			midi_fmout ^= TRUE;
			textcolor(CYAN);
			gotoxy(54,10);
			if(midi_fmout) vcprintf("ON  ");
			else vcprintf("OFF ");
		}
		else if(toupper(r) == 'U' && mpu_available) { // Toggle output
			midi_mpuout ^= TRUE;
			textcolor(CYAN);
			gotoxy(55,11);
			if(midi_mpuout) vcprintf("ON  ");
			else vcprintf("OFF ");
		}
	}
}


char vatoutstring[255];

/**************************************************************************
	void vcprintf(char *string,...)

	DESCRIPTION:  Special printf function that disables interrupts.  There is
								some concern that interrupting a printf may cause some
								systems to crash.  This is an experimental precaution.

**************************************************************************/
void vcprintf(char *string,...)
{
	va_list ap;

	_disable();

	va_start(ap, string);                  // sort out variable argument list
	vsprintf(vatoutstring,string,ap);      // dump output to a string
	cprintf("%s",vatoutstring);            // dump string to screen
	va_end(ap);                            // clean up
	_enable();
}
