/**************************************************************************
												VARMINT'S AUDIO TOOLS 0.61

	SOUND.C
	Written by: Eric Jorgensen (May, 1995)
								smeagol@rt66.com



	Written for: Turbo C++ V 3.1

		Compiler Settings:

		 - Stack checking should be turned off
		 - The default memory model for VAT is LARGE
		 - register variables should be turned off
		 - The setting for "assume SS == DS" should be "never".

		Linker Settings:

		 - "No stack" warning should be turned off

													 **** WARNING ****

 Use Varmint's Audio Tools at your own risk.  This code has only
 been lightly tested and has been found to cause cancer in laboratory rats.

**************************************************************************/



#include "sound.h"

//-------------------------------- Local Defs


#define WORD          unsigned int
#define BYTE          unsigned char
#define TD            midi_data->track[i] + midi_trackloc[i]
#define DOSYNC        if(sync_on && inportb(0x3da)&0x08) vsyncclock = 1000
#define DMAAUTOINIT   0x58
#define DMAONESHOT    0x48
#define VDB           strcat(vatdebugstring,vatapstr)
#define MAXSTOREDVOICES  32


//-------------------------------- Internal Function prototypes

WORD                  calcvibrato(WORD inc,WORD period,WORD amplitude);
int                   cardcheck(void);
static cardtype       checkhard(void);
void                  dmaset(BYTE far *sound_address,WORD len,BYTE channel);
void                  dodiv(void);
void                  dotick(void);
WORD                  downsemitone(WORD inc, WORD steps);
void                  enableint(BYTE nr);
void                  freeeffect(DWORD id);
void                  freesound(DWORD id);
void                  disableint(BYTE nr);
int                   getsbenv(void);
int                   getvoice(VOICE midi_voice[],int track,int channel, int note);
SOUNDEFFECT far       *grabsound(DWORD id);
BYTE                  int2vect(BYTE intnr);
WORD                  mcalc(WORD micro);
void _saveregs        mdelay(WORD delay);
void                  measure(void);
void                  midiplayer(void);
SOUNDEFFECT far       *newsoundeffect(void);
//WORD                  polldma(BYTE channel);
long int              readlong(FILE *infile);
int                   readshort(FILE *infile);
int                   readvarlen(BYTE *data,long int *value);
void far _saveregs _loadds interrupt sbint(void);
void                  sbremovevector(void);
void                  sbsetvector(void);
static int            scanint(void);
static int            testint(void);
WORD                  upsemitone(WORD inc, WORD steps);
void                  venable(void);
void                  vdisable(void);
cardtype              whichcard(void);


//-------------------------------- Global function pointers
																// These are used by the interrupt routines
																// To help keep track of interrupts that
																// Get shuffled around.

void far _saveregs _loadds interrupt  (*handlerint)(void) = sbint;
static void far interrupt (*orgint)(void) = NULL;
static void far interrupt (*orgtick)(void)= NULL;
static void far interrupt (*orgirqint)(void) = NULL;
static void far           (*call_func)(void);

//-------------------------------- FM variables and data

static WORD FM_off[9]={0,0x100,0x200,0x800,0x900,0xa00,0x1000,0x1100,0x1200};
static BYTE FM_fnr[12]={0x57,0x6b,0x81,0x98,0xb0,0xca,0xe5,0x02,0x20,0x41,0x63,0x87};
static BYTE FM_key_or[12]={1,1,1,1,1,1,1,2,2,2,2,2};
static BYTE FM_key[9];
static BYTE FM_keyscale1[9];
static BYTE FM_keyscale2[9];
static BYTE FM_vol[9] = {0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f,0x3f};

//-------------------------------- setup data and variables

static WORD ioaddr[6]={0x220,0x240,0x210,0x230,0x250,0x260};
static BYTE intrx[8]={5,7,2,3,10,11,12,15};
WORD        io_addr= 0x220;
WORD        intnr= 5;
WORD        dma_ch=1;
WORD        card_id=1;
WORD        fm_addr;
WORD        dsp_vers;
static WORD mue3,mue23,rythm=0xbd00;
static WORD dma_adr[8]= {0x00,0x02,0x04,0x06,0xc0,0xc4,0xc8,0xcc};
static WORD dma_len[8]= {0x01,0x03,0x05,0x07,0xc2,0xc6,0xca,0xce};
static WORD dma_page[8]={0x87,0x83,0x81,0x82,0x8f,0x8b,0x89,0x8a};
SBERROR     sberr = 0;
char        *errname[] = {
							"Cannot detect FMchip",
							"Cannot detect DSP",
							"Cannot find an open IRQ",
							"Cannot find an open DMA channel",
							"Cannot allocate memory for DMA buffer"};
int         tst_cnt;
BYTE        dma_controlbyte = DMAAUTOINIT;

//-------------------------------- DMA/DSP mixer varaibles and data

WORD        dma_bufferlen = 64;
BYTE far    *dma_buffer[2] = {NULL,NULL};
int         dma_currentbuffer = 0;
int far     *mix_buffer = NULL;
int         sounds_in_queue = 0;
WORD        sample_rate = 11000;
SAMPLE      *sounddata[MAXSOUNDS];
DWORD       soundpos[MAXSOUNDS];

SOUNDEFFECT far *soundlist = NULL;
SOUNDEFFECT far soundtoken[MAXSOUNDTOKENS];
DWORD       sampleid = 1;


//-------------------------------- Timer varaibles

static WORD timer_val=11000,timer_hold,timer_diff,mue999;
static WORD timadd,timsum;


//-------------------------------- Miscellaneous varaibles

long int        debugnum=0;
int         debug_lowdsp = FALSE;
int         debug_reverseflipflop = FALSE;
int         debug_forceflipflop = -1;
int         disabled = 0;
WORD        DSP_overhead = 0;

extern unsigned _stklen = 16000;  // The stack is usually only 4K, but
																	// I was getting stack overflow problems
																	// with that ammount.

WORD vsync_toolong = 4;           // This is the number of vclock ticks that
																	// tells us we have waited to long in
																	// the VarmintVSync() function.
DWORD       vclock=0;
WORD        vsyncclock = 0;
BYTE        sync_on = FALSE;
																	// This string is here to help debug
																	// startup problems
char        vatdebugstring[5000] = {"*** DEBUGGING STRING ***\n"};
char        vatapstr[255];

WORD midi_mpuport;

/**************************************************************************
	void GoVarmint(void)
	void DropDeadVarmint(void)

	DESCRIPTION:  Starts/stops the interrupt routine for Varmint's audio tools

**************************************************************************/
void GoVarmint(void)
{
	sbsetvector();                          // Install the sound kernel

	if(debug_lowdsp) dsp_vers = debug_lowdsp;

																					// Seems like later DSP versions
																					// need to have the flipflop in
																					// reverse order to remove static
	if(dsp_vers >= 0x300) debug_reverseflipflop = TRUE;
	if(debug_forceflipflop != -1) debug_reverseflipflop = debug_forceflipflop;


	if(dsp_vers >= 0x400) {                   // SB 4.0(SB16) or better?
                                          // Set up DMA channel.
		dmaset(dma_buffer[0],dma_bufferlen*2-1,dma_ch);
		DSPWrite(0xC6);                       // Generic DSP command
    	                                    // (8 bit, DAC, Auto-init)
    DSPWrite(0x00);												// Send Mode byte (mono,unsigned)
		DSPWrite((dma_bufferlen-1) & 0xff);   // Write length low byte
		DSPWrite((dma_bufferlen-1) >> 8);     // Write length high byte

	}
	else if(dsp_vers >= 0x300) {                   // SB Pro or better?

		DSPWrite(0x48);                       // Set DSP for 8bit Autoinit DMA
		DSPWrite((dma_bufferlen-1) & 0xff);   // Write length low byte
		DSPWrite((dma_bufferlen-1) >> 8);     // Write length high byte
		dmaset(dma_buffer[0],dma_bufferlen*2-1,dma_ch);

		DSPWrite(0x90);                       // Set DSP for High speed
																					// Autoinit 8bit DMA
	}
	else if(dsp_vers >= 0x200) {                   // SB2.0 or better?

		DSPWrite(0x48);                       // Set DSP for 8bit Autoinit DMA
		DSPWrite((dma_bufferlen-1) & 0xff);   // Write length low byte
		DSPWrite((dma_bufferlen-1) >> 8);     // Write length high byte
		dmaset(dma_buffer[0],dma_bufferlen*2-1,dma_ch);

		DSPWrite(0x1C);                       // Set DSP for Autoinit 8bit DMA
	}
	else {                                  // SB 1.x?
		dma_controlbyte = DMAONESHOT;
	}

	DSPWrite(DSP_INVOKE_INTR);              // Ignition!
}

void DropDeadVarmint(void)
{
	DSPWrite(0xD0);                         // Halt DMA
	if(dsp_vers >= 0x200) DSPWrite(0xDA);     // Halt Autoinitialized DMA
	DSPWrite(0xD0);                         // Halt DMA
	sbremovevector();                       // Clean up sound kernel
}


/*  ---------------  DSP  Stuff ------------ */




/**************************************************************************
	int DSPReset()

	DESCRIPTION: Resets the DSP

	RETURNS:
		TRUE or FALSE to indicate wether reset was sucsessful

**************************************************************************/
int DSPReset(void)
{
	int i;

	sprintf(vatapstr,"DSPReset() - <entry>\n"); VDB;

	mdelay(mue3);                         // Wait 3 microseconds
	outportb(io_addr+DSP_RESET,1);        // Write a 1 to the DSP reset port
	mdelay(mue3);                         // Wait 3 microseconds
	outportb(io_addr+DSP_RESET,0);        // Write a 0 to the DSP reset port


	for(i=0;i<50;i++) {                   // DSP should send back an 0xaa
		mdelay(mue3);
		if(DSPRead()==0xaa) return(TRUE);
	}

	return(FALSE);
}


/**************************************************************************
	BYTE DSPRead()

	DESCRIPTION:  reads a byte from the dsp

	RETURNS:
		The byte that was read

**************************************************************************/
BYTE DSPRead(void)
{
																			// Read until high bit of status port
																			// is set.
	while(!(inportb(io_addr+DSP_RSTATUS) &  0x80));
	return (inportb(io_addr+DSP_READ)); // Send back the value of the read port
}

/**************************************************************************
	void DSPWrite(BYTE output)

	DESCRIPTION: Writes a byte to the DSP

	INPUT:
		output  byte to write to the DSP

**************************************************************************/
void DSPWrite(BYTE output)
{
																			// Read until high bit of status port
																			// is clear.
	while((inportb(io_addr+DSP_WSTATUS) &  0x80));
	outportb(io_addr+DSP_WRITE,output); // Write our byte
}



/**************************************************************************
	WORD DSPGetVersion()

	DESCRIPTION:  Get the version number of the DSP

	RETURNS:
		DSP version number. High byte = major version, low byte = minor version

**************************************************************************/
WORD DSPGetVersion(void)
{
	DSPWrite(DSP_GET_VERS);
	return((WORD)DSPRead()*256+DSPRead());
}
/**************************************************************************
	int getsbenv()

	DESCRIPTION:  Get sound blaster information from the environment
								variable "BLASTER"

	RETURNS:
		TRUE if successful, FALSE if there was a problem

**************************************************************************/
int getsbenv(void)
{
	static char *envstr;                    // Environment strings must be
																					// static or global
	char str[255];
	int i;
	int outvalue = TRUE;

	sprintf(vatapstr,"getsbenv() - <entry>\n"); VDB;

	envstr=getenv("BLASTER");
	if(!envstr) return(FALSE);              // no blaster variable? go home
	strcpy(str,envstr);

	sprintf(vatapstr,"getsbenv() - Raw BLASTER=%s\n",str); VDB;

																					// Convert string to upper case
	for(i = 0 ; i < strlen(str); i++) *(str+i) = toupper(*(str+i));
																					// pick apart variable for info.
																					// Io address
	for(i = 0; *(str+i) != 0 && *(str + i) != 'A'; i++);
	if(*(str+i)){
		sscanf(str+i+1,"%x",&io_addr);
		if(io_addr<0x210 || io_addr>0x260) outvalue = FALSE;
	}
																					// MIDI port address
	for(i = 0; *(str+i) != 0 && *(str + i) != 'P'; i++);
	if(*(str+i)) sscanf(str+i+1,"%x",&midi_mpuport);

																					// Dma channel number
	for(i = 0; *(str+i) != 0 && *(str + i) != 'D'; i++);
	if(*(str+i)){
		sscanf(str+i+1,"%d",&dma_ch);
		if(dma_ch > 7) {
			outvalue = (FALSE);                 // only 0-7 allowed
			dma_ch = 1;                         // Set a reasonable default
		}
	}

																					// IRQ interrupt number
	for(i = 0; *(str+i) != 0 && *(str + i) != 'I'; i++);
	if(*(str+i)){
		sscanf(str+i+1,"%d",&intnr);
		if(intnr < 2  || intnr > 15) {
			outvalue =  (FALSE);
			intnr = 7;
		}
	}

																					// card_id
	for(i = 0; *(str+i) != 0 && *(str + i) != 'T'; i++);
	if(*(str+i)){
		sscanf(str+i+1,"%d",&card_id);
		if(card_id < 1 && card_id > 6) outvalue = FALSE; // 1 = SB 1.x, 6 = SB AWE
	}

	sprintf(vatapstr,"getsbenv() - Processed BLASTER=A%X P%X I%u D%u T%u\n",io_addr,midi_mpuport,intnr,dma_ch,card_id); VDB;

	return(outvalue);
}



/*  ---------------  Misc.  Stuff ------------ */



/**************************************************************************
	int cardcheck()

	DESCRIPTION:  Check for both FM chip and DSP

	RETURNS:
		Returns a value that has FM and DSP status information

**************************************************************************/
int cardcheck(void)
{
	int ret=0;

	ret|=FM_DETECT;
	if(DSPReset()) ret|=DSP_DETECT;
	return(ret);
}


/**************************************************************************
static void far _saveregs interrupt testnint()

	DESCRIPTION:  This function is stored as an interrupt to test
								various interrupt vectors by testint()

**************************************************************************/
static void far _saveregs interrupt testnint(void)
{
	tst_cnt++;                               // increment our test counter

	inportb(io_addr+DSP_RSTATUS);            // Acknowledge DSP interrupt

	outportb(0x20,0x20);                     // Clear PIC
	outportb(0xa0,0x20);
}



/**************************************************************************
	static int testint()

	DESCRIPTION:  This function is used by scanint() to test interrupt
								stuff.  It installs a test interrupt in the
								requested spot (intnr) then sees if the DSP can
								use it.

	RETURNS:
		TRUE if successful, FALSE if not

**************************************************************************/
static int testint(void)
{
	int i;
	BYTE int1,int2;

	orgint=_dos_getvect(int2vect(intnr));     // Save original interrupt
	tst_cnt=0;                                // reset our test interrupt counter.

	int1 = inportb(0x21);                     // Save PIC settings
	int2 = inportb(0xa1);

	vdisable();                               // Disable interrupts
	outportb(0x21,0xff);                      // Turn off all interrupts
	outportb(0xa1,0xff);

																						// put in our test interrupt
	_dos_setvect(int2vect(intnr),testnint);
	enableint(intnr);                         //  Turn on that interrupt
	venable();                                // Enable interrupts

	DSPWrite(DSP_INVOKE_INTR);                // Force DSP interrupt

	for(i=0;i<30000;i++) if(tst_cnt) break;   // wait for interrupt code to happen

	if(i == 30000) {                          // normalize things if the interrupt
																						// did not get called.
		inportb(io_addr+DSP_RSTATUS);           // Acknowledge DSP interrupt

		outportb(0x20,0x20);                    // Clear PIC
		outportb(0xa0,0x20);
	}

	vdisable();                               // disable inerrupts
	disableint(intnr);                        // put original interrupt back
	_dos_setvect(int2vect(intnr),orgint);

	outportb(0x21,int1);                      // restore PIC settings
	outportb(0xa1,int2);
	venable();                                // enable interrupts

	if(i==30000) return(FALSE);               // Timed out? No good!
	else return(TRUE);
}


/**************************************************************************
	static int scanint()

	DESCRIPTION: This makes sure that the interrupt number picked by the
								IRQ specification is a good choice.

	RETURNS:
		0 if failed, Interrupt number if successful
**************************************************************************/
static int scanint(void)
{
	int i;

	sprintf(vatapstr,"scanint() - <entry>\n"); VDB;

	if(testint()) return(intnr);   // Original choice good?

	sprintf(vatapstr,"scanint() - original interrupt failed (%d)\n",intnr); VDB;

	for(i=0;i<8;i++)                // Try our eight best guesses
	{
		intnr=intrx[i];
		sprintf(vatapstr,"scanint() - trying interrupt: %d\n",intnr); VDB;
		if(testint()) return(intnr);
	}

	sprintf(vatapstr,"scanint() - All interrupt tries failed\n"); VDB;

	return(0);
}


/**************************************************************************
	static cardtype checkhard()

	DESCRIPTION:  Checks hardware for DSP and FM chip

	RETURNS:
		The  detected Card Type.  (For now this is only an SB2.0)

**************************************************************************/
static cardtype checkhard(void)
{
	int ret;

	sprintf(vatapstr,"checkhard() - <entry>\n"); VDB;

	ret=DSPReset();

	sprintf(vatapstr,"checkhard() - DSPreset: %d\n",ret); VDB;

	if(ret)
	{
		if(!scanint()) {                    // Scan IRQ's
			sberr= irqerr;
			return(none);
		}
		sprintf(vatapstr,"checkhard() - scanint chose: %d\n",intnr); VDB;

		fm_addr=FM_ADLIB_ADDRESS;           // Check adlib chip first
		sprintf(vatapstr,"checkhard() - FM address: %X\n",fm_addr); VDB;

		return(sb20);
	}
	sberr = nodsperr;
	return(none);
}


/**************************************************************************
	cardtype whichcard()

	DESCRIPTION:  Calls various functions to make sure you've
								got a Sound Blaster.  Also checks to make sure that
								it has the right sound blaster environment settings.

	RETURNS:
		0 on failure, cardtype on success
**************************************************************************/
cardtype whichcard(void)
{
	cardtype cret = nodsp;
	int i;

	sprintf(vatapstr,"whichcard() - <entry>\n"); VDB;

	if(getsbenv()) cret=checkhard();      // grab environment variable
	if(cret!=nodsp) return(cret);         // If dsp is there, then go home

	sprintf(vatapstr,"whichcard() - Original settings failed\n"); VDB;


	for(i=0;i<6;i++)                      // scan around for a better io address
	{
		io_addr=ioaddr[i];
		sprintf(vatapstr,"whichcard() - Trying new io address: %X\n",io_addr); VDB;

		cret=checkhard();
		if(cret!=nodsp) return(cret);
	}
	sprintf(vatapstr,"whichcard() - FAILED COMPLETELY\n"); VDB;

	return(none);                         // Uh oh.
}


/**************************************************************************
	void SBCleanUp()

	DESCRIPTION: This conveniently cleans up all the stuff done by SBSetUp()
							 and other sound functions

	RETURNS:
		TRUE on success, FALSE on failure

**************************************************************************/
void SBCleanUp(void)
{
	DSPWrite(DSP_SPKR_OFF);

	farfree(dma_buffer[0]);
	farfree(mix_buffer);

}
/**************************************************************************
	int SBSetUp()

	DESCRIPTION: Sets up the sound blaster for action.  This is the only
							 function a programmer should really use.  Most of the
							 nitty gritty is handled internally.

	RETURNS:
		TRUE on success, FALSE on failure

**************************************************************************/
int SBSetUp(void)
{
	int i;

	InitTimerFunctions();         // Initialize counter 2 for timing functions

	sprintf(vatapstr,"Varmint's Audio Tools Version %s\n",VAT_VERSION); VDB;
	sprintf(vatapstr,"SBSetUp() - InitTimerFunctions\n"); VDB;
	sprintf(vatapstr,"SBSetUp() - measure [tval: %u] [ms: %u]\n",timer_val,mue999); VDB;

																// Calculate some special delays.
	mue3=mcalc(6) ;
	mue23=mcalc(46) ;

	sprintf(vatapstr,"SBSetUp() - mcalcs- [m3: %u] [m23: %u]\n",mue3,mue23); VDB;

																				// Go and check the hardware

	if(whichcard()==none) return(FALSE);

																				// Get DSP ready
	dsp_vers=DSPGetVersion();
	DSPWrite(DSP_SPKR_ON);

	sprintf(vatapstr,"SBSetUp() - DSP version: %d.%d%d\n",dsp_vers>>8,(dsp_vers&0xff)/10,(dsp_vers&0xff)%10); VDB;

																				// Allocate space for playback buffer
	dma_buffer[0] = (BYTE far *)farmalloc(dma_bufferlen*2+5);
	if(!dma_buffer[0]) {
		sberr = nomem;
		return(FALSE);
	}
	dma_buffer[1] = dma_buffer[0]+dma_bufferlen;

																				// Allocate space for Mixing buffer
	mix_buffer = (int far *)farmalloc(dma_bufferlen*2+10);
	if(!mix_buffer) {
		sberr = nomem;
		return(FALSE);
	}
																				// Clear the buffer
	for(i = 0; i <= dma_bufferlen; i++) {
		*(dma_buffer[0]+i) = 127;
		*(dma_buffer[1]+i) = 127;
	}
	SetSampleRate(11000);                 // Set the sample rate


	for(i = 0; i < MAXSOUNDTOKENS; i++){  // Set up sound tokens
		soundtoken[i].type = v_unused;
	}

	sprintf(vatapstr,"SBSetUp() - Processed BLASTER=A%X P%X I%u D%u T%u\n",io_addr,midi_mpuport,intnr,dma_ch,card_id); VDB;

	return(TRUE);
}


/**************************************************************************
	DWORD far2long(char far *adr)

	DESCRIPTION: This is used by dmaset to convert a regular far address
							 to a 20 bit flat address.

	RETURNS:
		The converted flat address
**************************************************************************/
DWORD far2long(char far *adr)
{
	return(((DWORD)FP_SEG(adr)<<4)+FP_OFF(adr));
}


/**************************************************************************
	void SetSampleRate(WORD rate)

	DESCRIPTION:  Sets the sample rate (specified in hz).

								Note:  The sample rate most likely will not be exactly
								equal to the rate specified in the argument argument since
								the Sound blaster can only accept rates in increments of
								400 hz.

	INPUTS:
		rate    playback rate in hertz

	RETURNS:
		The actual value of the new sample rate.

**************************************************************************/
WORD SetSampleRate(WORD rate)
{
	DWORD val;

	if(rate<4000) rate = 4000;          // 4000 is as slow as the SB can go
																			// Calculate number for the sound card
  if(rate>45000) rate = 45000;				// 45 Khz is the  praticle cutoff here.

	val=256-1000000L/rate;
	DSPWrite(DSP_SAMPLE_RATE);
	DSPWrite((BYTE)val);
																			// This helps other functions
																			// To know how fast the DSP is going
	sample_rate = (1000000L/(256-val));
	return sample_rate;
}

/**************************************************************************
	void dmaset(BYTE far *sound_address,WORD len,BYTE channel)

	DESCRIPTION:  This programs the DMA controller to start a single pass
								output transfer.

								(Draeden of VLA has provided some good information for
								DMA programming in his INTRO to DMA document)

	INPUTS:
		sound_address   Regular address of the DMA playback buffer
		len             length of the DMA buffer in bytes minus one
		channel         DMA  channel to use

**************************************************************************/
void dmaset(BYTE far *sound_address,WORD len,BYTE channel)
{
	WORD adr;
	DWORD adrl;
	BYTE page;

	adrl = far2long(sound_address);     // convert address to 20 bit format
	adr=(WORD)adrl;                     // extract page address
	page=(BYTE)(adrl>>16);              // extract page number

																			// PREPARE DMA.
																			//   (Channels 0-3 have different
																			//   command ports than 4-7.)

																			// SET CHANNEL, MASK BIT AND MODE

	if(channel <4) {                    // Chanels 0-3?
		outportb(0x0a,channel+4);          // write channel number with 3rd bit set
		outportb(0x0c,0);                  // Clear Byte Pointer

																			// Set the mode.  The mode is determined
																			// in GoVarmint().
		outportb(0x0b,dma_controlbyte+channel);
	}
	else {                              // channels 4-7
		outportb(0xd4,channel);            // write channel number
		outportb(0xd8,0);                  // Clear Byte Pointer

																			// Set the mode.  The mode is determined
																			// in GoVarmint().
		outportb(0xd6,dma_controlbyte+channel-4);
	}
																			// SET TRANSFER INFORMATION

	outportb(dma_adr[channel],adr&0xff);// Write address low byte
	outportb(dma_adr[channel],adr>>8);  // Write address high byte

	outportb(dma_len[channel],len&0xff);// Write length  low byte
	outportb(dma_len[channel],len>>8);  // Write length  high byte

	outportb(dma_page[channel],page);   // Write page

																			// CLEAR MASK BITS
	if(channel < 4) outportb(0x0a,channel);
	else outportb(0xd4,channel & 0x03);

}

/**************************************************************************
	void polldma(BYTE channel)

	DESCRIPTION:  This function poles the DMA controller to find out how many
								bytes are left in the current transfer.

								As of version 0.4, this function is no longer used, but
								I thought it might be useful to someone else, so I've only
								commented it out.
**************************************************************************/
/*WORD polldma(BYTE channel)
{
	BYTE low1,high1,low2,high2;

	disableint(intnr);             // Turn off the interrupt so we don't get
																	// caught with our pants down.
	asm {
		mov dx,0x0c                   // Flip the master reset switch
		mov al,0
		out dx,al
	}
	_DX = dma_len[channel];         // Load in the counter address
																	// read position twice, becasue sometimes
																	// there is a problem
	asm{
		in al,dx                      // read the low byte first
		mov low1,al
		in al,dx                      // read the high byte next
		mov high1,al

		in al,dx                      // read the low byte first
		mov low2,al
		in al,dx                      // read the high byte next
		mov high2,al

	}
	enableint(intnr);              // Done, so we'll put the interrupt back.

																	// High bytes the same? Use second reading
	if(high1 == high2)return((WORD)high2*256+low2);
																	// else the First reading is accurate
	return(high1*256+low1);

}*/



/**************************************************************************
	void far _saveregs interrupt sbint()

	DESCRIPTION:  This is the sound Blaster interrupt that is to be
								called at the end of DMA transfer.  This is how the flow
								of things goes:

									Mix regular samples
									Mix MOD stuff
									Calculate next DMA half of buffer
									acknowledge interrupt
									Run the midi stuff.

								The DOSYNC's are macros define in sound.h designed to
								check the vertical retrace and then set vsyncclock so that
								these sound functions can be used with games that check
								the vertical retrace.

**************************************************************************/
void far _saveregs _loadds interrupt sbint(void)
{
	int i,j,mcount,mskip;
	WORD lastbyte;
	int far *mbuf;
	BYTE far *dbuf;
	SOUNDEFFECT far *s,*hold;

	vdisable();                      // disable interrupts();

	if(DSP_overhead) TimerOn();      // start timer to measure DSP overhead

	DOSYNC;                          // Macro for checking the vertical retrace
																	 // clear out the mixing buffer
	for(j = 0; j < dma_bufferlen+1; j++) *(mix_buffer+j) = 128;
	DOSYNC;
																	 // ***********************************
																	 // *                                 *
																	 // *       Mix in plain sounds       *
																	 // *                                 *
																	 // ***********************************
	s = soundlist;
	while(s){
		hold = s->next;
		if(s->active && s->type == v_plain) {
			if(*s->pos >= dma_bufferlen) {// enough left to fill the buffer?
				for(j = 0; j < dma_bufferlen; j++) *(mix_buffer+j) += *(s->data++);
				*s->pos -= (dma_bufferlen);
			}
			else {                         // else play what is left
				for(j = 0; j <= *s->pos; j++) *(mix_buffer+j) += *(s->data++);

				if(s->chain) {               // Is there a sound to play after this?
					(s->chain)->next = soundlist;
					soundlist = s->chain;
					s->chain = NULL;           // Break link to hide from freesound()
					sounds_in_queue++;
				}
				freesound(s->id);            // delete the sound
			}
		}
		DOSYNC;
		s = hold;
	}
	DOSYNC;

																	 // ***********************************
																	 // *                                 *
																	 // *      Mix in fancy sounds        *
																	 // *                                 *
																	 // ***********************************
	s = soundlist;
	while(s){
		hold = s->next;
		if(s->active && s->type == v_fancy) {
			for(j = 0; j < dma_bufferlen; j++) {
																		// Add this byte to the mixing buffer
				*(mix_buffer+j) += (*(s->data + *s->pos) * s->volume)>>4;
																		// Here is where the frequency magic
																		// happens.  I use fixed point arithmatic
																		// to increment the sample position pointer
																		// at different rates.
				s->position += s->pinc;
																		// Are we at the end of the sample?
				if(*s->pos >= s->length) {
					if(!s->repeat) {          // Do we not repeat this sample?

						if(s->chain) {          // Is there an effect to play after this?
							(s->chain)->next = soundlist;
							soundlist = s->chain;
							s->chain = NULL;      // Break link to hide from freesound()
							sounds_in_queue++;
						}

						freesound(s->id);      // Terminate it if not.
						break;
					}
					else {                    // Decrememnt the repeat counter
						if(s->repeat > 0) s->repeat--; // -1 = infinity
						s->position = 0;
					}
				}
			}
		}
		DOSYNC;
		s = hold;
	}
	DOSYNC;

	if(debug_reverseflipflop) dma_currentbuffer ^= 1;


																	 // Write mixed stuff to the playback buffer
	for(mbuf = mix_buffer, dbuf = dma_buffer[dma_currentbuffer];
			mbuf != mix_buffer+dma_bufferlen;
			mbuf++) {
		if(*mbuf > 254) *mbuf = 254;       // Clip overload values
		else if(*mbuf < 1) *mbuf = 1;
		*(dbuf++) = *mbuf;
	}
	DOSYNC;
	if(!debug_reverseflipflop) dma_currentbuffer ^= 1;

	if(dsp_vers < 0x200) {                 // Not SB 2.0?
																					// Set up DMA for one-shot
		dmaset(dma_buffer[dma_currentbuffer],dma_bufferlen-1,dma_ch);
		DOSYNC;
		DSPWrite(0x14);                       // One shot DMA DAC (8bit)
		DSPWrite((dma_bufferlen-1) & 0xff);  // Write length low byte
		DSPWrite((dma_bufferlen-1) >> 8);    // Write length high byte
	}


	DOSYNC;


																	 // Varmint's system clock
	vclock ++;                       // Vertical retrace sync clock
	vsyncclock ++;

	if(DSP_overhead) DSP_overhead = TimerOff();//  How long did this take?

	inportb(io_addr+DSP_RSTATUS);    // Acknowledge DSP interrupt

	outportb(0x20,0x20);             // Clear PIC
	outportb(0xa0,0x20);

	venable();                       // Enable interrupts again
}



/**************************************************************************
	void sbsetvector()

	DESCRIPTION:  Installs the DMA interrupt vector.  This makes it so that
								sbint() is called whenever a DMA transfer is finished

**************************************************************************/
void sbsetvector(void)
{
	orgirqint=_dos_getvect(int2vect(intnr));
	_dos_setvect(int2vect(intnr),handlerint);     /* set vector to our routine */
	enableint(intnr);                             /* enable sb interrupt */
}


/**************************************************************************
	void sbremovevector()

	DESCRIPTION:  Removes the DMA interrupt vector

**************************************************************************/
void sbremovevector(void)
{
	disableint(intnr);                            /* disable sb interrupt */
	_dos_setvect(int2vect(intnr),orgirqint);      /* restore org intr vector */
}


/* --------------------------------------------------- */

/* timerX routines are following

	These routines are for highly accurate time measurements

*/


/**************************************************************************
	void InitTimerFunctions()

	DESCRIPTION: Initializes speaker timer for timing operations.

**************************************************************************/
void InitTimerFunctions(void)
{
	BYTE b;

	b = inportb(0x61) & 0xfd;          // Read current port value.

																		 // Disable speaker output, but keep
																		 // timer 2 enabled.
	outportb(0x61,b | 0x01);

	measure();                         // Setup for delay functions

}


/**************************************************************************
	void TimerOn()

	DESCRIPTION: Turns on timer counter for a time measurement

**************************************************************************/
void TimerOn(void)
{
	outportb(0x43,0xb0);                 // Set up the timer for countdown
	outportb(0x42,0xff);                // Least significant byte
	outportb(0x42,0xff);                // Most significant byte
}


/**************************************************************************
	WORD TimerOff()

	DESCRIPTION: Turns off time and reports clicks elapsed.  Note that this
							 timer is so quick that it is wraps after only 56
							 milliseconds.  If you want to timer longer stuff, I suggest
							 using the global variable vclock.  It's tick frequency is
							 sample_rate / dma_bufferlen.

							 To convert to micro seconds, multiply the return value by
							 0.838.

	RETURNS:
		Current timer count

**************************************************************************/
WORD TimerOff(void)
{
	outportb(0x43,0xc0);                 // Read the timer countdown status
	timer_hold = inportb(0x42) + inportb(0x42) * 256;

	timer_diff = 0xffff - timer_hold;   // timer_diff is used later
	return timer_diff;
}



/**************************************************************************
	void measure()

	DESCRIPTION: measures a standard delay loop for other delay functions

**************************************************************************/
void measure(void)
{
	vdisable();                       // Disable interrupts

	TimerOn();                        // Start the timer

	asm mov cx,10000                  /* internal test loop */
loop1:
	asm loop loop1

	timer_val = TimerOff();           // read the timer

	venable();                        // Enable interrupts

	mue999 = mcalc(990);              // Calculate a millisecond
}


/**************************************************************************
	void mdelay(WORD delay)

	DESCRIPTION: Very tiny delay

	INPUT:
		Length of delay in processor counts

**************************************************************************/
void _saveregs mdelay(WORD delay)
{
	asm mov cx,delay
loop1:
	asm loop loop1
}


/**************************************************************************
	void _saveregs MilliDelay(WORD delay)

	DESCRIPTION:  Millisec delay.  When using this library, you should
								use this delay for millisecond delays instead of the delay
								function that comes with turbo C.

	INPUTS:
		delay       Length of delay in millisconds

**************************************************************************/
void _saveregs MilliDelay(WORD delay)
{
	WORD i;

	for(i=0;i<delay;i++) mdelay(mue999); // mdelay(mue999) = 1 millisec
}

/**************************************************************************
	void _saveregs MicroDelay(WORD delay)

	DESCRIPTION:  Microsec delay.

								Note: This function probably has around a 10-15
								microsecond overhead, so I would only recommend it for
								delays bigger than 50 microseconds and smaller than 5
								milliseconds (5000 microseconds).

	INPUTS:
		delay       Length of delay in microsconds

**************************************************************************/
void _saveregs MicroDelay(WORD delay)
{
	mdelay(mcalc(delay));
}


/**************************************************************************
	WORD mcalc(WORD micro)

	DESCRIPTION:  Calculates number of ticks to send to mdelay for a specified
	number of microseconds.

	INPUTS:
		micro   Number of microseconds

	Returns:
		Value to send to mdelay() that will generate the specified number of
		microseconds.

**************************************************************************/
WORD mcalc(WORD micro)
{
	DWORD d;

	d = ((DWORD)micro*11940/timer_val);
																			//This clip is to prevent wrap around
																			// That can occur when a computer is
																			// too fast or the argument
																			// is too big.
	if(d > 0xffff) d = 0xffff;
	return((WORD)d);
}

																			/********************************

																								TIMER 0 STUFF




																			*********************************/
/* --------------------------------

CAUTION: These routines can cause a lot of headaches while debugging.
If you set your own interrupt and then stop the program before you call
RemoveTimer0(), you'd better reboot your computer, because very
unpredictable things will happen if InstallTimer0() is called again.
My suggestion is to get your interrupt working and then comment out the
Timer0 routines until the rest of the program is written and debugged.
																			- Eric  */

/**************************************************************************
	static void interrupt timerint()

	DESCRIPTION: THis is the actual interrupt function stored in the timer
							 0 slot (int 08).   This calls the old int08 function
							 at proper intervals as well as the user specified function

**************************************************************************/
static void interrupt timerint(void)
{

	TimerOn();                      // set timer for overhead calculation.
	call_func();                     // user specified function

																	 // Now let's do some fancy counting so
																	 // we can call the system clock at the
																	 // right moments.
	if(timsum<100)
	{
		timsum+=timadd;
		orgtick();
	}
	else
	{
		asm mov al,0x20;
		asm out 0x20,al;
	}
	timsum-=100;                     // decrement our special timer
}


/**************************************************************************
	void InstallTimer0(WORD period,void far (*func)())

	DESCRIPTION: This sets up timer0 to call your function at the specified
							 period.

	INPUTS:
		period    period in system clock ticks.  (1 tick = 0.838 microsecs)
		func      pointer to the function to install

**************************************************************************/
void InstallTimer0(WORD period,void far (*func)(void))
{
	 if(!func) return;       /* no valid func ptr */
	call_func=func;

	timadd= (WORD)(6553600L/period); // counting seed for timerint()
	timsum=0;                 // start counter at 0

	asm mov al,0x36            /* program timer 0 with modus 3 */
	asm out 0x43,al            /* and counter value of period  */
	asm mov ax,period
	asm out 0x40,al
	asm mov al,ah
	asm out 0x40,al

	orgtick= _dos_getvect(8);       // Remember the old interrupt
	_dos_setvect(8,timerint);      // put in a new one.
}


/**************************************************************************
	void RemoveTimer0()

	DESCRIPTION:  Removes your goofy interrupt, 'cause we didn't want
								it anyway!  :P

**************************************************************************/
void RemoveTimer0(void)
{
	if(!orgtick) return;      // Must have called InstallTimer0 first
	asm mov al,0x36            /* program timer 0 with modus 3 */
	asm out 0x43,al            /* and counter value of 0 (2^16)*/
	asm mov al,0
	asm out 0x40,al
	asm out 0x40,al
	_dos_setvect(8,orgtick);       // put back original vector
}



/**************************************************************************
	BYTE int2vect(BYTE intnr)

	DESCRIPTION:  This function converts a PIC irq number to a true
								interrupt vector number.  For PC's with a 286 or greater,
								irq's 0-7  refer to interrupts 0x08 - 0x0F and
								irq's 8-15 refer to interrupts 0x70 - 0x77.

	INPUTS:
		intnr   IRQ number

	RETURNS:
		Actual vector corresponding to the IRQ

**************************************************************************/
BYTE int2vect(BYTE intnr)
{
	if(intnr>7) return(intnr + 0x68);
	else return(intnr+8);
}


/**************************************************************************
	void enableint(BYTE nr)

	DESCRIPTION:  Enables an IRQ interrupt using the Programmable
								interrupt controller (PIC)

								To enable a interrupt, you want to turn its PIC bit off.
								Let's say that you want to turn on interrupt 3.  The
								code works something like this:


										BIT:                 76543210

										1 << nr             00001000
										~(1 << nr)          11110111 (mask)
																				01001101 (setting)
										setting & mask      01000101
																						^
																						|
													Notice how bit 3 is set to zero.

	INPUTS;
		nr    IRQ number

**************************************************************************/
void enableint(BYTE nr)
{
	BYTE mask,setting;

	if(nr<8) {                             // First controller?
		mask = ~(1 << nr);                  // Create interrupt mask
		setting = inportb(0x21) & mask;     // Turn off the proper bit
		outportb(0x21,setting);             // Write back the result
	}
	else {                                // Second Controller?
		mask = ~(1 << (nr-8));              // Create interrupt mask
		setting = inportb(0xa1) & mask;     // Turn off the proper bit
		outportb(0xa1,setting);             // Write back the result
	}
}
void testenableint(BYTE nr,BYTE *test)
{
	BYTE mask,setting;

	mask = ~(1 << nr);                  // Create interrupt mask
	setting = (*test) & mask;     // Turn off the proper bit
	*test = setting;             // Write back the result
}


/**************************************************************************
	void disableint(BYTE nr)

	DESCRIPTION:  Disables an IRQ interrupt using the Programmable
								interrupt controller.

								To disable a interrupt, you want to turn its PIC bit on.
								Let's say that you want to turn off interrupt 3.  The
								code works something like this:


										BIT:                 76543210

										1 << nr             00001000
										~(1 << nr)          11110111 (mask)
																				01000101 (setting)
										~setting            10111010
										(~setting) & mask    10110010 (new value of setting)
										~setting            01001101

																						^
																						|
													Notice how bit 3 is set to one.
	INPUTS;
		nr    IRQ number

**************************************************************************/
void disableint(BYTE nr)
{
	BYTE mask,setting;

	if(nr<8) {                            // First controller?
		mask = ~(1 << nr);                  // Create interrupt mask
		setting = (~inportb(0x21)) & mask;  // Turn on the proper bit
		outportb(0x21,~setting);            // Write back the result
	}
	else {                                // Second Controller?
		mask = ~(1 << (nr-8));              // Create interrupt mask
		setting = (~inportb(0xa1)) & mask;  // Turn on the proper bit
		outportb(0xa1,~setting);            // Write back the result
	}
}
void testdisableint(BYTE nr,BYTE *test)
{
	BYTE mask,setting;

	mask = ~(1 << nr);                  // Create interrupt mask
	setting = (~(*test)) & mask;  // Turn on the proper bit
	*test = ~setting;            // Write back the result
}





/**************************************************************************
	void freesound(DWORD id)

	DESCRIPTION:  Removes sound effect from active list and clears token for
								later use.

	INPUTS:
		id    Id of sound effect to free

**************************************************************************/
void freesound(DWORD id)
{
	SOUNDEFFECT far *s,*hold,*hold2;


	hold = NULL;                       // Find the sound based on the id
	s = soundlist;
	while(s && s->id != id) {          // move down until we find the id
		hold = s;
		s = s->next;
	}

	if(!s) return;                     // No such sound? go home!

	sounds_in_queue--;                 // keep track for PlaySound()
	s->type = v_unused;
	s->id = 0;

	if(!hold) {                        // Hold empty? Must be head of list
		soundlist = s->next;             // Close the list
		return;
	}

	hold->next = s->next;              // close the spot in the list

	hold = s->chain;                   // Kill any chains attached to it
	while(hold) {
		hold2 = hold->chain;
		hold->type = v_unused;
		hold->id = 0;
		hold = hold2;
	}
	return;
}


/**************************************************************************
	SOUNDEFFECT far *newsoundeffect(void)

	DESCRIPTION:  Creates a new sound effect an initializes it variables

								Note: Soundeffect Tokens are pre-allocated rather than dynamically
											allocated because many thousands of sounds may
											be played in a single game.  Allocating and
											re-allocating all this tiny pieces of memory can cause
											problems.

	RETURNS:
		Pointer to the new sound effect structure

	SAMPLE  *data;
	DWORD   length;
	DWORD   position;
	WORD    *pos;
	WORD    pinc;
	int     repeat;
	BYTE    volume;
	BYTE    active;
	DWORD   id;
	struct soundeffect *chain;
	struct soundeffect *next;

**************************************************************************/
SOUNDEFFECT far *newsoundeffect(void)
{
	SOUNDEFFECT far *s;
	int i;
																	 // Find an empty sound token
	for(i =0 ; i< MAXSOUNDTOKENS && soundtoken[i].type != v_unused; i++);
	if(i == MAXSOUNDTOKENS) return(NULL);
	s = &soundtoken[i];
																	 // Initialize varaibles
	s->data = NULL;
	s->length = 1;
	s->position = 0;
	s->pos = (WORD *)(((BYTE *)&(s->position))+ 1);
	s->pinc = 256;
	s->repeat = 0;
	s->volume = 16;
	s->active = TRUE;
	s->id = sampleid++;
	s->chain = NULL;
	s->next = NULL;
	s->chainfunc = NULL;
	return(s);
}

/**************************************************************************
	DWORD PlaySound(SAMPLE *data,DWORD length,VATPLAYTYPE type)

	DESCRIPTION: Adds an sound to the play list.  If the playlist is full,
							 the oldest one gets chopped;

	INPUTS:
		data      pointer to sample data
		length    length of sample in bytes
		type      inidicates which code is used to play the sound

	RETURNS:
		id of created sound effect or NULL on failure.  This id can is used
		as a handle to access the effect by other functions.


**************************************************************************/
DWORD PlaySound(SAMPLE *data,DWORD length,VATPLAYTYPE type)
{
	SOUNDEFFECT far *s;

	_disable();                               // We don't want any interrupts
																						// While we do this.

	if(sounds_in_queue >= MAXSOUNDS) {        // Too many sounds active?
																						// Chop off sound at end of list
		s = soundlist;
		while(s && s->next && s->next->repeat != -1 ) s = s->next;
		freesound(s->id);                       // Kill the sound
	}

	s = newsoundeffect();                     // Allocate new effect
	if(!s) {
		_enable();                              // restore interrupts
		return(0);
	}

	s->data = data;                           // Initialize data portion
	s->length = length-1;
	if(type == v_plain) *s->pos = (WORD)s->length;
	s->type = type;

	s->next = soundlist;                      // Add to the list
	soundlist = s;

	sounds_in_queue++;

	_enable();                                // retore interrupts
	return(s->id);
}


/**************************************************************************
	DWORD ChainSoundEffect(SAMPLE *data,length)

	DESCRIPTION:  This chains a new soundeffect to an active one.  The new
								sound effect wiill be played when the first one os finished.

	INPUTS:
		data      pointer to sample data
		length    length of sample in bytes
		type      playtype for the sew sample (v_fancy or v_plain)
		id        id if the sound effect to chain to

	RETURNS:
		id of created sound effect or NULL on failure.  This id can is used
		as a handle to access the effect by other functions.


**************************************************************************/
DWORD ChainSoundEffect(SAMPLE *data,DWORD length,VATPLAYTYPE type,DWORD id)
{
	SOUNDEFFECT far *s,*o;

	_disable();                               // We don't want any interrupts
																						// While we do this.

	o = grabsound(id);                       // grab the original effect
	if(!o) {
		_enable();                             // Restore interrupts
		return(0);
	}

	s = newsoundeffect();                    // Allocate new effect
	if(!s) {
		_enable();                             // restore interrupts
		return(0);
	}

	s->data = data;                          // Initialize data portion
	s->length = length-1;
	if(type == v_plain) *s->pos = (WORD)s->length;
	s->type = type;
																					 // Link the new soundeffect
	while(o->chain) o = o->chain;            // Find the end of the chain
	o->chain = s;

	_enable();                               // Restore interrupts
	return(s->id);
}


/**************************************************************************
	int AddChainFunction(DWORD id,void (*func)(SOUNDEFFECT far *))

	DESCRIPTION:  Attaches a function to a sound effect.  The function is
								called when the sound effect is done.

	INPUTS:
		id      Id of sound effect to use
		func    pointer to the function to chain to

	RETURNS:
		TRUE if successful, FALSE if not

**************************************************************************/
int AddChainFunction(DWORD id,void (*func)(SOUNDEFFECT far *))
{
	SOUNDEFFECT far *s;

	s = grabsound(id);              // find the soundeffect
	if(!s) return(FALSE);

	s->chainfunc = func;            // add the function

	return(TRUE);                   // It worked!
}

/**************************************************************************
	SOUNDEFFECT far *grabsound(DWORD id)

	DESCRIPTION:  Finds a sound structure matching the id

	INPUTS:
		id      id of the sound structure to find

**************************************************************************/
SOUNDEFFECT far *grabsound(DWORD id)
{
	int i;
																// Search sound tokens for matching id
	for(i = 0; i < MAXSOUNDTOKENS && soundtoken[i].id != id; i++);
	if(i == MAXSOUNDTOKENS) return NULL;

	return(&soundtoken[i]);       // Return what we got
}

/**************************************************************************
	void AlterSoundEffect(DWORD id,VATCOMMAND c,long int parameter)

	DESCRIPTION:  Alters items in the sound effect structure

	INPUTS:
		id          soundeffect id
		c           The command id to signal the appropriate action
		parameter   The argument of the command

**************************************************************************/
void AlterSoundEffect(DWORD id,VATCOMMAND c,long int parameter)
{
	SOUNDEFFECT far *s;

	s = grabsound(id);

	if(!s) return;

	switch(c) {                            // Act on all the varmint commands
		case v_play:
			s->position = 0;
			s->active = TRUE;
			break;
		case v_stop:
			freesound(id);
			break;
		case v_pause:
			s->active = FALSE;
			break;
		case v_resume:
			s->active = TRUE;
			break;
		case v_rewind:
			s->position = 0;
			break;
		case v_setrepeat:
			s->repeat = parameter;              // positive value = loop x times
																					// negative value = loop forever
																					// 0 = loop once
			break;
		case v_setpos:
			*s->pos = parameter;
			break;
		case v_setrate:                       // Rate is sent in as a fixed point
			s->pinc = parameter;
			break;
		case v_setvolume:
			s->volume = parameter;              // 32 = LOUD, 0 = silent
			break;
		default:
			break;
	}

}





/*********************************************************************
	SAMPLE *LoadWave(char *wavefile,unsigned long int *length)

	DESCRIPTION:  Loads a wave files  (mono, 8bit)

	INTPUTS:

		wavefile  filename of wave file
		length    pointer to length interger

	RETURNS:
		pointer to data

**********************************************************************/
SAMPLE *LoadWave(char *wavefile,unsigned long int *length)
{
	int i;
  WORD ii;
	SAMPLE  *data;
	BYTE dummydata[255];
	FILE *input;
	DWORD rlen,flen;
	WORD s_per_sec,b_per_sec,num_channels,tag;
	char riffid[5],waveid[5],fmtid[5],dataid[5];

	input = fopen(wavefile,"rb");
	if(!input) {                              // If unsuccesful...
		*length = 1;                             // set short length to prevent
																						// mistakes later.
		return(NULL);                           // REturn a null pointer
	}
																						// Get WAVE header data

	fread(riffid,1,4,input);                  // wave files staqrt with "Riff"
	riffid[4] = 0;
	fread(&rlen,1,4,input);                   // File size
	fread(waveid,1,4,input);                  // Wave id string  ("Wave")
	waveid[4] = 0;
	if(strcmp(waveid,"WAVE")) {               // is it a wave file?
		fclose(input);
		return(NULL);
	}

	fread(fmtid,1,4,input);                   // Format id string ("fmt ")
	fmtid[4] = 0;
	fread(&flen,1,4,input);                   // offset to data
	if(flen > 240) flen = 240;                // Just a precaution so that
																						// We do not overload dummydata

	fread(&tag,1,2,input);                    // tag
	fread(&num_channels,1,2,input);           // number of channels
	fread(&s_per_sec,1,2,input);              // sample rate (hz)
	fread(&b_per_sec,1,2,input);              // bytes per seconf rate
	fread(dummydata,1,(size_t)flen-8,input);          // Skip ahead
	fread(dataid,1,4,input);                  // Dataid string
	dataid[4] = 0;
	fread(length,1,4,input);                  // length of data

	data = (SAMPLE *)farmalloc(*length+1);    // allocate memory for data
	if(!data) {                                // oops.  Not enough mem!
		fclose(input);
		return(NULL);
	}

	fread(data,1,(size_t)*length,input);              // read the data

	for(ii = 0; ii < *length; ii++) {            // convert to signed format
		*(data + ii) = ((BYTE)*(data + ii))-128;
	}

	fclose(input);                            // Wrap it up
	return(data);
}

/**************************************************************************
	void VarmintVSync(void)

	DESCRIPTION: This vsync function pays attention to vsyncclock so that we
								 never miss a vsync.  The vsync can be missed if sbint
								gets called while we are waiting for the retrace.

								The variable vsync_toolong should be roughly equalt to
								the number of times you expect sbint to get called between
								vertical retraces.  The will be afffected by dma_bufferlen,
								sample_rate, and the frequency of the VGA monitor

**************************************************************************/
void VarmintVSync(void)
{
	while(!(inportb(0x3da)&0x08)) {     // Wait for retrace to start
		if(vsyncclock>vsync_toolong) break;   // Have we waited too long?
	}

	while(inportb(0x3da)&0x08);          // Wait for retrace to finish
	vsyncclock = 0;                     // reset the clock
}


/**************************************************************************
	void TimeVSync(void)

	DESCRIPTION:  This times a vertical retrace and sets the variable
								vsynctoolong to an appropriate value based on the sample
								rate and dma_bufferlen.

								This should be called before GoVarmint();

**************************************************************************/
void TimeVSync(void)
{
	WORD l;

	VarmintVSync();                // Wait for a Vync to end

	TimerOn();                    // Time the next vsync
	VarmintVSync();
	l = TimerOff();

	if(l < 100) l = 17000;         // Sanity measure

																 // Figure how many vclock ticks will occur
																 // During a retrace.
	vsync_toolong = (WORD)((long)l * sample_rate / 1193000L / dma_bufferlen) + 2;
	//vsync_toolong = l  / 1193000.0 * sample_rate/dma_bufferlen + 1.5;
}


/**************************************************************************
	WORD PercentOverhead(WORD ticks)

	DESCRIPTION:  Calculates the Percentage CPU overhead that the
								blaster interrupt consumes.

	INPUTS:
		ticks       Number of clock ticks that pass during the interrupt

	RETURNS:
		An interger percentage (* 10) based on dma_bufferlen and sample rate.
		ie:  0 = 0% overhead, 1000 = 100.0% overhead


**************************************************************************/
WORD PercentOverhead(WORD ticks)
{
	//return ((ticks / 1193000.0)* (sample_rate / dma_bufferlen) * 100.0);
	return((DWORD)ticks * sample_rate / (1193L * dma_bufferlen));
}


/**************************************************************************
	void vdisable(void)

	DESCRIPTION:  Special interrupt disable function for VAT

**************************************************************************/
void vdisable(void)
{
	if(disabled) {                 // Don't do anything if interrupts are
		disabled++;                  // already disabled.
		return;
	}

	_disable();                    // disable the interrupts
	disabled++;
}

/**************************************************************************
	void venable(void)

	DESCRIPTION:  Special interrupt enable function for VAT

**************************************************************************/
void venable(void)
{
	disabled--;

	if(!disabled) _enable();       // enable the interrupts if this is the
																 // outermost enable.
}
