#include <stdio.h>
#include <stdlib.h>
#include "math.h"
#include "vwcolors.h"

#define UNEXPOSED 0
#define MAXEXPOSURE 10000
#define TRUE 1
#define FALSE 0

typedef enum { BUTTON = 1, FIELD, AREA, PLAQUE, RADIO, SPINNER, SLIDER} WINTYPE;
typedef enum { ONCE, CONTINUOUS } SLIDERTYPE;
typedef enum { PLAIN = 1,STRETCH,CLIP,CENTER,TILE} ICONFILL;

#define WINTYPE_BUTTON 1
#define WINTYPE_FIELD 2
#define WINTYPE_AREA 3
#define WINTYPE_PLAQUE 4
#define WINTYPE_RADIO 5
#define WINTYPE_SPINNER 6
#define WINTYPE_SLIDER 7
#define WINTYPE_VALUEPIT 8
#define WINTYPE_LABEL 9
#define THICKBORDERWIDTH 3
#define THINBORDERWIDTH 2
#define FIELDBORDERWIDTH 2
#define BUTTONBORDERWIDTH 1
#define PLAQUEBORDERWIDTH 3
#define LABELBORDERWIDTH 2


#define CALLOC(x,y)  (y *)calloc(x+1,sizeof(y))
#define MALLOC(x) (x *)malloc(sizeof(x))

typedef struct window  {
	int x,y,w,h;                		// geometry
	int ret;                        // return value
	int exposure;                   // exposure level
  int *intpointer;                // data pointer for radio buttons
	long int intdata[5];            // internal data
	int hotn;                       // hot character strinng position
	char hotc;                      // hot character value
	unsigned char tc,thc,hc,sc,bc;  // colors (text,texthighlight,highlight,shadow,background)
	char *stringdata;               // internal text data
	char *text;                     // text to be displayed on button
	char *name;                     // name used to get window handle
	int type;                       // type of window
	int dump_on_press;							// flag for mouseloop()
	void *background;               // pointer for backing storage
	void *icon;                     // pointer to icon;
	ICONFILL icon_fillstyle;        // determines appearance of icon
	void (*callback)();             // callback for automatic handling
	struct window *next;            // link to next window in llist
	struct window *child,*sibling,*parent;  // used for grouping windows
} WINDOW;

																		// function prototypes
void			initwindows(void);
int 			readwindows(char *winfilename);
void 			drawwindow(char *);
WINDOW 		*getwindow(char *);
int				getwindowid(char *);
void 			DrawBox(WINDOW *win,int);
void 			DrawPit(WINDOW *win,int);
int 			inwindow(int x, int y);
void 			exposewindow(char *win);
void 			DrawSpinner(WINDOW *win);
void 			DrawRadio(WINDOW *win);
void 			hidewindows(void);
void 			hidewindow(char *win);
void 			errorhandler(int i, char *string);
char 			inkey(void);
void 			locate(int,int);
void 			setwintext(WINDOW *win, char *string);
void 			setwincolors(char *win,int tc,int thc,int hc,int sc,int bc);
void 			blankwindow(char *win,int color);
int  			getrelease(int *x, int *y);
int 			getpress(int *x, int *y);
int 			getsliderstatus(char *win,int x, int y);
void 			updateslider(char *win,int val);
void 			DrawSlider(WINDOW *w);
void 			setintdata(char *win,int index,long int number);
int 			toggleradio(char *win);
void 			getinput(int x,int y,int fc,int bc, char *prompt,char *string,int maxlength);
WINDOW 		*inwhatwindow(int x, int y);
void 			buttonaction(WINDOW *win, int action, int bw);
int 			prepslider(char *textwin, char *swin);
WINDOW 		*addwindow(char *name,int x,int y,int w,int h,int type,char *text);
void 			drawpit(int x1,int y1,int x2,int y2,int bw,int hc, int sc, int bc);
void 			drawbox(int x1,int y1,int x2,int y2,int bw,int hc, int sc, int bc);
void 			setslidervals(char *sliderwin, long int val,long int low,
										long int high,long int inc,void (*callback)());
void 			setcallback(char *window,void (*callback)());
WINDOW 		*getwinfromid(int id);
int 			mouseloop(int *x, int *y);
void 			setchild(char *parent, char *child);
long int 	doslider(char *sliderwin);
void 			setradio(char *win,int *pointer,int bitnum);


																	 // globals :(
extern char errstring[256];
extern WINDOW *windows;
extern WINDOW *wtest;
extern int graphics_on;
extern WINDOW *debugwin;

													/* External primitives: define these functions in
														 your own modules using the graphics library
														 of your choice. */
void VW_hidemouse(void);
void VW_showmouse(void);
void VW_mousestatus(int *x, int *y, int *buttons);
void VW_bar(int color,int x1,int y1,int x2,int y2);
void VW_line(int color,int x1,int y1,int x2,int y2);
void VW_circle(int color,int x1,int y1, int radius);
void VW_floodfill(int color,int x,int y, int border);
void VW_outtextxy(int color,int x, int y, char *string);
int  VW_textwidth(char *);
int	 VW_textheight(char *);